/*****************************************************************
Counter Firmware for Geiger Mueller counter. Uses 8x2 Char LCD
to display output. Timer0 and tmr0_msb variable are used to count
events during a one second intervall, which are stored in counter.
The number of counts per second are stored in an array, later summed
and multiplyed by 6 to get the CPM.

PD3 (INT1) is used to change between counting modes. Counting modes 
include per minute and accumulated. Counting input is PD4 (T0).

Project: GM Counter
MCU Type: ATtiny2313
Clock: 8,00MHz

Data Types:

int8_t - Signed Char 
uint16_t - Unsigned Int 
uint32_t - Unsigned Long 
int64_t - Signed Long Long

(signed/unsigned) char - 1 byte 
(signed/unsigned) short - 2 bytes 
(signed/unsigned) int - 2 bytes 
(signed/unsigned) long - 4 bytes 
(signed/unsigned) long long - 8 bytes 
float - 4 bytes (floating point) 
double - alias to float

******************************************************************/

// Flag definitions
#define READY 0
#define ACCUM 1
#define CPM 2
#define RELAT 3

// USART definitions
#define USART_BAUDRATE 38400
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "lcd.h"

// Declare global variables
volatile uint8_t tmr0_msb;             			// increments with every timer 0 overflow
volatile uint16_t counter;             			// Total number of events captured in 1 second
volatile uint8_t flag_reg;             			// Register for flags used in program

// Find Counts per minute
uint16_t count_math(void)
{
static uint16_t old_cnt[10];             		// Array used to store previous count values
static uint8_t j;								// Used to run through array
uint8_t i;
uint16_t cntpermin;

// Place counts during last second in array
j++;
if (j>=10) j=0;
old_cnt[j] = counter;
		
// Sum up counts over latest 10 second period
cntpermin = 0;
for (i=0; i<10; ++i) cntpermin += old_cnt[i];
			
// Multiply with 6 to get CPM
cntpermin *= 6;

return cntpermin;
}

// Write value to LCD
void display(uint16_t data)
{
char line_txt[8];

// Formatt output
utoa(data, line_txt, 10);

// Write to LCD
lcd_gotoxy(0, 1);
lcd_puts(line_txt);
}
// Timer 0 overflow interrupt service routine
ISR(TIMER0_OVF_vect)
{
    tmr0_msb++;

}

// Timer 1 overflow interrupt service routine
ISR(TIMER1_OVF_vect)
{
// Disable Interrupts
	cli();
    
// Find total number of events during last second
    counter = (255 * tmr0_msb) + TCNT0;
	
	flag_reg |= (1 << READY);
	
// Set Timer 1 start point for 1 second interrupt rate
    TCNT1H=0xE17B >> 8;
    TCNT1L=0xE17B & 0xff;
	
	// Reinitialize Timer values, Enable Interrupts
	tmr0_msb = 0;
    TCNT0 = 0;
	sei();

}

// External Interrupt Rutine, used to change metering modes
ISR(INT1_vect)
{
// Toggle mode
flag_reg ^= (1 << CPM);
flag_reg ^= (1 << ACCUM);
}

// Initialize USART Module
void USART_Init(void)
{

	//Enable Transmitter
	UCSRB |= (1 << TXEN);
	
	// Load lower 8-bits of the baud rate value into the low byte of the UBRR register 
	UBRRL = BAUD_PRESCALE;
	// Load upper 8-bits of the baud rate value into the high byte of the UBRR register 
	UBRRH = (BAUD_PRESCALE >> 8);
	
	//Set data frame format: asynchronous mode,no parity, 1 stop bit, 8 bit size
	// This is set by default in ATtiny2313, so not really necessary
	UCSRC = (0<<UMSEL)|(0<<UPM1)|(0<<UPM0)|(0<<USBS)|(1<<UCSZ1)|(1<<UCSZ0);
	
}

// Send a byte over USART
void USART_SendByte(uint8_t data)
{
// Wait if a byte is being transmitted
while((UCSRA&(1<<UDRE)) == 0);

// Transmit LSBs of data
UDR = data;
}

// Send Formatted Data through UART
void formatted_serial(uint16_t data)
{
	// Send Start Bytes
	USART_SendByte(255);
	USART_SendByte(255);
	USART_SendByte(255);
	// Send MSB
	USART_SendByte(data >> 8);
	// Send LSB
	USART_SendByte(data);
	// Send End Byte
	USART_SendByte(128);
}

// Main
void main(void)
{
uint16_t temp = 0;

PORTD = 0;

// Timer/Counter 0 initialization, set clock source to T0 pin rising Edge
TCCR0B = (1<<CS02)|(1<<CS01)|(1<<CS00);

// Timer/Counter 1 initialization
// Set Prescaler to 1:1024, set to interrupt evey second
TCCR1B = (1<<CS12)|(0<<CS11)|(1<<CS10);
TCNT1H = 0xE1;
TCNT1L = 0x7B;

// Timer 0 and 1 Overflow Interrupt initialization
TIMSK = (1<<TOIE1)|(1<<TOIE0);

// Turn off analog comparator to save power
ACSR = (1<<ACD);

// INT1 External Interrupt enable
GIMSK = (1<<INT1);
// ...on INT1 falling edge
MCUCR = (1<<ISC11)|(0<<ISC10);

// Initialize UART
USART_Init();

// LCD module initialization
lcd_init(LCD_DISP_ON);
lcd_clrscr();

// Set default mode
flag_reg |= (1 << CPM);

sei();

while(1)
{
	// Simple check so display in only updated once per second
	if (bit_is_set(flag_reg, READY))
	{
	// New count data, display and send out via UART
	formatted_serial(counter);
	
		// Display accumulated dose
		if (bit_is_set(flag_reg, ACCUM))
		{
			temp += counter;
			// Write to LCD;
			lcd_clrscr();
			lcd_puts("Tot.Dose");
			display(temp);
			flag_reg  &=~(1 << READY);
		}
		// Display counts per minute
		else if (bit_is_set(flag_reg, CPM))
		{
			// Write to LCD
			lcd_clrscr();
			lcd_puts("cnt/min");
			display(count_math());
			flag_reg  &=~(1 << READY);
		}
		
	}
}


}